import math

from pyramid.response import Response

from pyramid_restful.pagination import PageNumberPagination
from pyramid_restful.pagination.utilities import replace_query_param


class JSONAPIPagination(PageNumberPagination):
    """
    Page number pagination for JSON API formatted responses.
    """

    max_page_size = 50
    page_query_param = 'page[number]'
    page_size_query_param = 'page[size]'

    def get_paginated_response(self, data):
        links = data.get('links', {})
        first_url = self.get_first_link()
        last_url = self.get_last_link()

        if not data['data'] or first_url == last_url:  # Don't include links if there is only a single page.
            return Response(json=data)

        next_url = self.get_next_link()
        previous_url = self.get_previous_link()

        if next_url:
            links['next'] = next_url
        if previous_url:
            links['prev'] = previous_url
        if first_url:
            links['first'] = first_url
        if last_url:
            links['last'] = last_url

        data['links'] = links

        meta = data.get('meta', {})
        meta['count'] = self.page.paginator.count
        data['meta'] = meta

        return Response(json=data)

    def get_first_link(self):
        url = self.request.current_route_url()
        return replace_query_param(url, self.page_query_param, 1)

    def get_last_link(self):
        url = self.request.current_route_url()
        count = self.page.paginator.count
        page_size = self.get_page_size(self.request)
        total_pages = int(math.ceil(count / float(page_size)))
        return replace_query_param(url, self.page_query_param, total_pages)


class IncludeRelationshipsMixin:
    """
    Parses query strings params for related objects that should be included in the serialized response.
    If includable_relationships is None then all relationships can be included.

    includable_relationships = A dictionary where the key is the mapped querystring value and the value is a dict of:
        {'rel': '<relationship name from schema>',
        'join': <optional, sqla join method to use>,
        'options': <optional, sqla query options>}
    """

    QUERY_KEY = 'include'
    includable_relationships = None

    def get_schema(self, *args, **kwargs):
        includable_names = self.includable_relationships.keys() if self.includable_relationships else None
        includes = []

        for key, vals in self.request.params.items():
            if key == self.QUERY_KEY:
                for val in vals.split(','):  # Allow for a comma separated list of include values
                    if includable_names is None or val in includable_names:
                        includes.append(val)

        if includes:
            kwargs['include_data'] = includes

        return super(IncludeRelationshipsMixin, self).get_schema(*args, **kwargs)

    def get_query(self):
        """
        Allows for the query to be dynamically updated based on any included relationships.
        Allowing for data to be pre-joined or pre-fetched cutting down on the number of db
        queries required for the request.
        """

        query = super(IncludeRelationshipsMixin, self).get_query()
        includables = getattr(self, 'includable_relationships', [])

        if includables:
            requested_includes = []

            for key, val in self.request.params.items():
                if key == self.QUERY_KEY:
                    requested_includes.extend(val.split(','))

            if requested_includes:
                available_includes = self.includable_relationships.keys()

                for name in requested_includes:
                    if name in available_includes:
                        field = self.includable_relationships[name]

                        if field.get('join'):
                            query = getattr(query, field.get('join'))(field['rel'])

                        # Apply optional options
                        options = field.get('options')

                        if options:
                            query = query.options(*options)

        return query


class NestableSchemaMixin:
    """
    I forget where I left off on this but I think this will only work for a single level of nesting.

    Used when you need to use a Nested field with a JSONAPI Schema.
    It formats the errors so that the nested schemas errors pointers have the correct attribute path.

    For this to work, if a nested schema needs to have many=True, set it on an instance of the schema, not
    as a Nested argument (you will get a Invalid Type error).

    GOOD: fields.Nested(MySchema(many=True), required=True)
    BAD:  fields.Nested(MySchema,many=True, required=True)
    """

    def format_nested_errors(self, key, errors):
        """
        Updates the pointer attribute in an errors object so that it points to the nested attribute.

        :param key: The nested field name.
        :param errors: An array of already jsonapi formatted errors.
        :return: The formatted errors array.
        """

        for error in errors:
            source = error['source']
            source['pointer'] = '/data/attributes/{}'.format(key) + source['pointer']

        return errors

    def format_errors(self, errors, many):
        if not errors:
            return {}
        if isinstance(errors, (list, tuple)):
            return {'errors': errors}

        formatted_errors = []

        if many:  # won't ever be a nested jsonapi schema
            for index, errors in errors.items():
                for field_name, field_errors in errors.items():
                    formatted_errors.extend(
                        [self.format_error(field_name, message, index=index) for message in field_errors])
        else:
            for field_name, field_errors in errors.items():
                if isinstance(field_errors, dict):
                    errors = field_errors.get('errors')

                    if not errors:
                        # Not JSON API formatted, happens when a required nested attribute does not exist in the data
                        errors = self.format_errors(field_errors, False)['errors']

                    formatted_errors.extend(self.format_nested_errors(field_name, errors))
                else:
                    formatted_errors.extend([self.format_error(field_name, message) for message in field_errors])

        return {'errors': formatted_errors}



from marshmallow import class_registry
from marshmallow.base import SchemaABC
from marshmallow_jsonapi.fields import Relationship
from marshmallow.compat import basestring

_RECURSIVE_NESTED = 'self'


class IncludableRelationship(Relationship):
    """
    Allows an alternative attribute to be specified for include relationships.

    This allows you to specify a normal foreign key attribute when an relationship is
    not included and a relationship when the data should be included.

    Hopefully this will be PR'd into marshmallow_jsonapi in the near future and this can go away.
    """

    def __init__(self, include_attribute=None, *args, **kwargs):
        self.include_attribute = include_attribute
        super(IncludableRelationship, self).__init__(*args, **kwargs)

    @property
    def schema(self):
        """
        Overriding this to add the context of the
        """
        context = self.parent.context if self.parent else {}

        if isinstance(self.__schema, SchemaABC):
            return self.__schema
        if isinstance(self.__schema, type) and issubclass(self.__schema, SchemaABC):
            self.__schema = self.__schema(context=context)
            return self.__schema
        if isinstance(self.__schema, basestring):
            if self.__schema == _RECURSIVE_NESTED:
                parent_class = self.parent.__class__
                self.__schema = parent_class(include_data=self.parent.include_data, context=context)
            else:
                schema_class = class_registry.get_class(self.__schema)
                self.__schema = schema_class(context=context)
            return self.__schema
        else:
            raise ValueError(('A Schema is required to serialize a nested '
                              'relationship with include_data'))

    def _serialize(self, value, attr, obj):
        dict_class = self.parent.dict_class if self.parent else dict
        ret = dict_class()
        self_url = self.get_self_url(obj)
        related_url = self.get_related_url(obj)

        if self_url or related_url:
            ret['links'] = dict_class()
            if self_url:
                ret['links']['self'] = self_url
            if related_url:
                ret['links']['related'] = related_url

        # resource linkage is required when including the data
        if self.include_resource_linkage or self.include_data:
            if value is None:
                ret['data'] = [] if self.many else None
            else:
                ret['data'] = self.get_resource_linkage(value)

        if self.include_data:
            if self.include_attribute:
                value = getattr(obj, self.include_attribute, None)

            if value is not None:
                if self.many:
                    for item in value:
                        self._serialize_included(item)
                else:
                    self._serialize_included(value)

        return ret
